/*

    w32event.c - utility functions to log application event
    Copyright (C) 2004  Vesely Alessandro

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/**
* @file
* functions to log application event.
*/
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <w32event.h>

#include <stdio.h>
#include <stdarg.h>
#include <wchar.h>
#include <ctype.h>

#if !defined(PATH_MAX)
#if defined(MAX_PATH)
#define PATH_MAX MAX_PATH
#else
#define PATH_MAX 4096
#endif
#endif

#if defined(TEST_MAIN)
#undef MAX_INSERTIONS
#define MAX_INSERTIONS 8
#endif

// naive error logging
#if defined(NDEBUG)
#define LOCATION_DECL extern void unreferenced_f(void)
#define LOCATION_SET(descr) (void)0
#define LOCATION_PRINT(func) (void)0
#else
#define LOCATION_DECL char const*location = NULL
#define LOCATION_SET(descr) location = (descr)
#define LOCATION_PRINT(func) location_print((func), error, location)
void location_print(char const *func, int err, char const *location);
#endif

static char *check_full_path(char *file, char *buf, unsigned *length)
// initially length is the length of the buffer
// return NULL on error, otherwise set length to the actual name length
{
   char *last;
   unsigned len;

   if (strchr(file, '\\') != NULL || strchr(file, '/') != NULL)
   {
      *length = strlen(file) + 1;
      return file;
   }

   len = SearchPath(
      NULL,               // default application search path
      file,               // application name
      ".dll",             // default extension
      *length,            // size of buffer
      buf,                // buffer
      &last);             // last component (in buffer)

   if (len == 0 || len >= *length)
      return NULL;

   *length = len + 1;
   return buf;
}

/**
* Register an event source.
* Create or update the key
* <code>
* HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\<i>pszSrcName</i>
* </code>.
*
* This function is called as part of the initialization,
* so it should be called only when doing custom initialization.
*
* @param pszSrcName event source name
* @param pszMsgDLL name or path for message file (<code>.dll</code>
*   extension is used by default. The path is searched if the name
*   contains no slashes.)
* @param dwNum number of categories defined in the message file.
* @return a handle that can be used to call <code>ReportEvent</code>
*   or NULL in case of failure.
*
* @see log_init()
* @see @ref msdn3
*/
HANDLE SetEventSource(LPTSTR pszSrcName,
   LPTSTR pszMsgDLL, DWORD  dwNum)
{
   HKEY hk;
   DWORD dwData;
   DWORD error;
   int hk_is_open = 0;
   char szBuf[PATH_MAX];
   char fname[PATH_MAX];
   char *dll_name;
   unsigned dll_name_size;
   HANDLE rtc;
   DWORD flags;

   LOCATION_DECL;

   // check if path is given
   dll_name_size = sizeof fname;
   dll_name = check_full_path(pszMsgDLL, fname, &dll_name_size);
   if (dll_name == NULL)
   {
      error = GetLastError();
      LOCATION_SET("check_full_path");
      goto func_exit;
   }

   // Create the event source as a subkey of the log.

   sprintf(szBuf,
      "SYSTEM\\CurrentControlSet\\Services\\EventLog\\%s\\%s",
      "Application", pszSrcName);

   if ((error =
      RegCreateKeyEx(HKEY_LOCAL_MACHINE, // subkey handle
         szBuf,                          // subkey to open or create
         0,                              // must be 0
         NULL,                           // class can be null
         REG_OPTION_NON_VOLATILE,        // options
         KEY_WRITE,                      // access rights mask
         NULL,                           // security attributes
         &hk,                            // result handle
         NULL)) != 0)                    // don't care if created or not
   {
      LOCATION_SET("RegCreateKeyEx");
      goto func_exit;
   }

   hk_is_open = 1;

   // Set the name of the message file.

   if ((error = RegSetValueEx(hk,   // subkey handle
      "EventMessageFile",           // value name
      0,                            // must be zero
      REG_EXPAND_SZ,                // value type
      (LPBYTE) dll_name,            // pointer to value data
      dll_name_size)) != 0)         // length of value data
   {
      LOCATION_SET("RegSetValueEx EventMessageFile");
      goto func_exit;
   }

   // Set the supported event types.

   dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE |
        EVENTLOG_INFORMATION_TYPE;

   if ((error = RegSetValueEx(hk,   // subkey handle
      "TypesSupported",             // value name
      0,                            // must be zero
      REG_DWORD,                    // value type
      (LPBYTE) &dwData,             // pointer to value data
      sizeof(DWORD))) != 0)         // length of value data
   {
      LOCATION_SET("RegSetValueEx TypesSupported");
      goto func_exit;
   }

   // Set the category message file and number of categories.

   if ((error = RegSetValueEx(hk,   // subkey handle
      "CategoryMessageFile",        // value name
      0,                            // must be zero
      REG_EXPAND_SZ,                // value type
      (LPBYTE) dll_name,            // pointer to value data
      dll_name_size)) != 0)         // length of value data
   {
      LOCATION_SET("RegSetValueEx CategoryMessageFile");
      goto func_exit;
   }

   if ((error = RegSetValueEx(hk,   // subkey handle
      "CategoryCount",              // value name
      0,                            // must be zero
      REG_DWORD,                    // value type
      (LPBYTE) &dwNum,              // pointer to value data
      sizeof(DWORD))) != 0)         // length of value data
   {
      LOCATION_SET("RegSetValueEx CategoryCount");
      goto func_exit;
   }

   rtc = RegisterEventSource(NULL, pszSrcName);
   if (rtc == NULL)
   {
      LOCATION_SET("RegisterEventSource");
      error = GetLastError();
      goto func_exit;
   }

#if !defined(NDEBUG)
   if (GetHandleInformation(rtc, &flags))
   {
      printf("The log handle is%s inheritable\n",
         (flags & HANDLE_FLAG_INHERIT) ? "" : " not");
   }
   else
   {
      LOCATION_SET("GetHandleInformation");
      error = GetLastError();
   }
#else
   (void)flags;
#endif

   func_exit:
   {
      if (hk_is_open)
         RegCloseKey(hk);
      if (error)
      {
         SetLastError(error);
         LOCATION_PRINT("SetEventSource");
      }
   }


   return rtc;
}

static PSID dup_sid(PSID orig, char const*const where)
{
   PSID sid = NULL;
   DWORD error = 0;
   DWORD sid_length = GetLengthSid(orig);

   LOCATION_DECL;

   if (sid_length <= 0 || sid_length > 4096)
   {
      LOCATION_SET("GetLengthSid");
      error = ERROR_INVALID_SID;
      goto func_exit;
   }

#if !defined(NDEBUG)
   printf("got SID from %s, length = %u\n", where, sid_length);
#else
   (void)where;
#endif

   sid = (PSID)malloc(sid_length);
   if (sid == NULL)
   {
      LOCATION_SET("malloc");
      error = ERROR_NOT_ENOUGH_MEMORY;
      goto func_exit;
   }

   memset(sid, 0, sid_length);
   if (!CopySid(sid_length, sid, orig))
   {
      LOCATION_SET("CopySid");
      error = GetLastError();
      free(sid);
      sid = NULL;
   }

   func_exit:
   {
      if (error)
      {
         SetLastError(error);
         LOCATION_PRINT("dup_sid");
      }
   }

   return sid;
}

static void *get_token(HANDLE h, TOKEN_INFORMATION_CLASS req)
{
   void *info = NULL;
   DWORD rtc_length = 0;
   DWORD error;

   LOCATION_DECL;

   GetTokenInformation(h,                   // token handle
      req,                                  // ask for a TOKEN_OWNER
      NULL,                                 // output buffer
      0,                                    // size we can take
      &rtc_length);                         // returned length
   if ((error = GetLastError()) != ERROR_INSUFFICIENT_BUFFER)
   {
      LOCATION_SET("GetTokenInformation");
      goto func_exit;
   }
   error = 0;

   if ((info = malloc(rtc_length)) == NULL)
   {
      LOCATION_SET("malloc");
      error = ERROR_NOT_ENOUGH_MEMORY;
      goto func_exit;
   }
   memset(info, 0, rtc_length);

   if (!GetTokenInformation(h,              // token handle
      req,                                  // ask for a TOKEN_OWNER
      info,                                 // output buffer
      rtc_length,                           // size we can take
      &rtc_length))                         // returned length
   {
      error = GetLastError();
      LOCATION_SET("GetTokenInformation");
      free(info);
      info = NULL;
      goto func_exit;
   }

   func_exit:
   {
      if (error)
      {
         SetLastError(error);
         LOCATION_PRINT("get_token");
      }
   }

   return info;
}

/**
* Obtain a SID for logging.
* This functions tries to obtain a user SID, an owner SID
* or the groups logon SID, in that order. If it fails
* returns NULL.
*
* This function is called as part of the initialization,
* so it should be called only if doing custom initialization.
*
* @return a PSID for logging.
*
* @see log_init()
* @see @ref msdn3
*/
PSID GetEventSid()
{
   HANDLE h;
   int h_is_open = 0;
   DWORD error = 0;
   PSID sid = NULL;

   LOCATION_DECL;

   TOKEN_USER *token_user = NULL;
   TOKEN_OWNER *token_owner = NULL;
   TOKEN_GROUPS *token_groups = NULL;

   if (!OpenProcessToken(GetCurrentProcess(),
      TOKEN_QUERY | TOKEN_QUERY_SOURCE | TOKEN_READ, &h))
   {
      LOCATION_SET("OpenProcessToken");
      error = GetLastError();
      goto func_exit;
   }

   h_is_open = 1;

   if ((token_user = (TOKEN_USER*)get_token(h, TokenUser)) != NULL &&
      token_user->User.Sid != NULL &&
      IsValidSid(token_user->User.Sid))
   {
      sid = dup_sid(token_user->User.Sid, "USER Sid");
   }
   else if ((token_owner = (TOKEN_OWNER*)get_token(h, TokenOwner)) != NULL &&
      token_owner->Owner != NULL &&
      IsValidSid(token_owner->Owner))
   {
      sid = dup_sid(token_owner->Owner, "OWNER Sid");
   }
   else if ((token_groups = (TOKEN_GROUPS*)get_token(h, TokenGroups)) != NULL)
   {
      DWORD i;

      for (i = 0; i < token_groups->GroupCount; ++i)
      {
         SID_AND_ATTRIBUTES *const group = &token_groups->Groups[i];
         if ((group->Attributes & SE_GROUP_LOGON_ID) == SE_GROUP_LOGON_ID)
         {
            if (group->Sid != NULL && IsValidSid(group->Sid))
               sid = dup_sid(group->Sid, "Logon SID of GROUPS");
            break;
         }
      }
   }

   func_exit:
   {
      if (h_is_open)
         CloseHandle(h);
      free(token_user);
      free(token_owner);
      free(token_groups);
      if (error)
      {
         SetLastError(error);
         LOCATION_PRINT("GetEventSid");
      }
   }

   return sid;
}

/**
* Deallocate the SID that was created by GetEventSid().
* @param sid the SID that was created by GetEventSid()
* @see GetEventSid()
*/
void ClearEventSid(PSID sid)
{
   free(sid);
}

static char *windows_syserr(DWORD err)
/* result must be freed with LocalFree() */
{
   LPVOID lpMsgBuf = NULL;
   FormatMessage(
       FORMAT_MESSAGE_ALLOCATE_BUFFER |
       FORMAT_MESSAGE_FROM_SYSTEM |
       FORMAT_MESSAGE_IGNORE_INSERTS,
       NULL,
       err,
       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language */
       (LPTSTR) &lpMsgBuf,
       0,
       NULL
   );
   return lpMsgBuf;
}

static char *alloc_sys_message(int err, char const *defl)
/*
* result must be freed with free() unless it is defl
*/
{
   char buf[64], *p;
   char *msg = err ? windows_syserr(err) : NULL;
   unsigned len = msg ? strlen(msg) : 0;
   unsigned elen = sprintf(buf, "[%d]", err);
   char *rtc = (char*)malloc(len + elen + 2);

   if (msg)
   {
      if (rtc)
         strcpy(rtc, msg);
      LocalFree(msg);
   }

   if (rtc == NULL)
      return (char*)defl;

   for (p = rtc + len - 1; p >= rtc && isspace(*(unsigned char*)p); --p)
      *p = 0;
   if (p < rtc) p = rtc;
   else if (p > rtc) *p++ = ' ';
   strcpy(p, buf);
   return rtc;
}

#if !defined(NDEBUG)
void location_print(char const *func, int err, char const *location)
{
   static char const * const defl = "NO ERROR MESSAGE";
   char *msg = alloc_sys_message(err, defl);
   printf("w32event: %s: error in %s: %s\n", func, location, msg);
   if (msg != defl) free(msg);
}
#endif

/*
* already typedef'd in .h */
struct log_message_table
{
   unsigned nentries;
   struct entry_set_index
   {
      unsigned first_entry, last_entry, offset;
   } index[1]; // actually nentries
};

typedef struct log_message_entry
{
   unsigned short length;
   unsigned short is_unicode;
   unsigned char text[1]; // possibly unicode
} log_message_entry;


typedef struct parsed_fmt
{
   unsigned mask_M; // args that need replacing with message
   unsigned mask_P; // args that will be printed to file
   int count, ins_count;
   va_list args;
   LPCTSTR ins[MAX_INSERTIONS + 1]; // a NULL terminator is still needed
   char *fmt, *defl;
   unsigned fmt_len;
} parsed_fmt;

/**
* Internal argument separator.
* This sequence is used internally to separate formats given to
* fprintf. It should never be part of data. (If that happens,
* the insertions string will be split (thus becoming two)
* and the last valid insertion is discarded.)
*/
#define ARG_SEPARATOR "\x08\x01\x04\x02\x03\x07\x05"
#define ARG_SEPARATOR_LENGTH (sizeof (ARG_SEPARATOR) - 1)

static log_message_entry*search_message(log_event_data const *log, unsigned id)
{
   unsigned i, j;
   if (log && log->table)
   {
      log_message_table const *const table = log->table;
      char *ptr = (char*)table;
      unsigned const nentries = table->nentries;
      for (i = 0; i < nentries; ++i)
      {
         struct entry_set_index const *const ndx = &table->index[i];
         if (ndx->first_entry <= id && ndx->last_entry >= id)
         {
            ptr += ndx->offset;
            for (j = ndx->first_entry; j < id; ++j)
               ptr += ((log_message_entry*)ptr)->length;
            return (log_message_entry*)ptr;
         }
      }
   }
   return NULL;
}


static int simple_argument(parsed_fmt *pf, int ch)
/*
* Set a simple format, %s or %M, followed by ARG_SEPARATOR or
* retrieve the actual argument and place it in ins.
* Return 0 if successful, -1 on error.
*/
{
   int const count = pf->count++;

   if (pf->mask_P == 0) // set the argument now
   {
      int const ins_count = pf->ins_count++;
      if (ins_count < MAX_INSERTIONS)
         pf->ins[ins_count] = (ch == 'M') ?
            alloc_sys_message(va_arg(pf->args, int), pf->defl) :
            va_arg(pf->args, char*);
   }
   else // will fprintf this arg later
   {
      static const unsigned fmt_len = ARG_SEPARATOR_LENGTH + 2;
      char const *const fmt = (ch == 'M') ?
         "%d" ARG_SEPARATOR : "%s" ARG_SEPARATOR;
      if (pf->fmt_len > fmt_len)
      {
         pf->fmt_len -= fmt_len;
         strcpy(pf->fmt, fmt);
         pf->fmt += fmt_len;
      }
   }

   if (ch == 'M')
      pf->mask_M |= 1 << count;
   return 0;
}

static int reformat(parsed_fmt *pf, log_message_entry const *en)
/*
* Copy the original message string doing rough unicode conversion.
* Set the number of arguments and possibly set the corresponding value
* via simple_argument. Arguments that are not understood are flagged mask_P.
*/
{
   unsigned const is_unicode = (en->is_unicode & 1) != 0;
   unsigned const advance =  is_unicode ? sizeof(wchar_t) : sizeof (char);
   unsigned char const *src = &en->text[0];
   unsigned char const *esrc = ((unsigned char*)en) + en->length;

   while (src < esrc && pf->count < MAX_INSERTIONS)
   {
      int ch = is_unicode ? *(wchar_t*)src : *src;
      src += advance;
      if (ch == '%')
      {
         unsigned char const *start = src - advance;
         if (src < esrc)
         {
            ch = is_unicode ? *(wchar_t*)src : *src;
            if (ch == '%')
            {
               src += advance; // skip "%%"
               continue;
            }
            if (ch == 's' || ch == 'M')
            {
               src += advance;
               simple_argument(pf, ch);
               continue;
            }
         }

         while (src < esrc) // reach end of complex format
         {
            ch = is_unicode ? *(wchar_t*)src : *src;
            src += advance;
            if (ch == 0 || ch == '%' || ch == 'n')
            {
               start = NULL;
               src -= advance; // restart, e.g. for "% blah %.8s"
               break;          // %n should not be used
            }
            if (strchr("diouxXfeEgGcCsSp", ch) != NULL)
               break;
         }

         if (start != NULL && src < esrc)
         {
            char *fmt = pf->fmt;
            unsigned len = pf->fmt_len;
            while (len > 0 && start < src)
            {
               --len;
               *fmt++ = (char)(is_unicode ? *(wchar_t*)start : *start);
               start += advance;
            }
            if (len > ARG_SEPARATOR_LENGTH)
            {
               strcpy(fmt, ARG_SEPARATOR);
               pf->fmt_len = len - ARG_SEPARATOR_LENGTH;
               pf->fmt = fmt + ARG_SEPARATOR_LENGTH;
               pf->mask_P |= 1 << pf->count;
            }
            else *pf->fmt = 0;
            pf->count += 1;
         }
      }
   }
   return 0;
}

/**
* Log the event using a <code>va_list</code>.
*
* This is the workhorse function. It arranges the parameters for
* <a href="http://msdn.microsoft.com/library/en-us/debug/base/reportevent.asp"
* target="_blank"><code>ReportEvent</code></a>.
*
* The category is read from log_event_data. If it is less than 1,
* then the facility number will be used, thus implementing the
* <i>category == facility</i> schema. If the category is higher than
* the last_category, then the latter will be used.
*
* It retrieves the formal argument definition and rewrites it
* so as to exclude spurious characters (typically newlines).
* It recognizes a sequence <b><code>\%...<i>T</i></code></b> where
* the terminating character <code><i>T</i></code> is one of
* <code>diouxXfeEgGcCsSp</code>, and <b><code>...</code></b> is one
* or more formatting flags. (Note that <code>\%n</code>
* is not supported.)
*
* The special sequence
* <b><code>\%M</code></b> is recognized. The corresponding
* integer argument is replaced with the system message string
* that can be retrieved via <code>FormatMessage</code> using the
* default language. The actual error number is appended to the
* message.
*
* For simple formats, <code>\%s</code> and <code>\%M</code>,
* the function
* directly substitutes the corresponding parameter. Otherwise, it rewrites
* the format string inserting an <code>ARGUMENT_SEPARATOR</code> after
* the terminating character.
* Then it calls <code>vfprintf</code> on a temporary file, passing all
* the remaining parameters.
* Finally, it reads back the temporary file and builds the array
* of insertions as required.
*
* Up to MAX_INSERTIONS parameters are accepted.
*
* That way, callers can seamlessly pass any parameter that printf would
* accept, including 64 bits data. All parameterts are converted to strings
* according to the format specified in the message file. Note that NT can
* make a savvy use of temporary files (most probably it will never reach
* the hard disk) therefore you only pay a negligible performance penality
* for this. However, under pathological circumstance this can fail:
*
* @warning If the app has too many files opened and tries to report a failure
* using formats other than a combination of <code>\%s</code> and
* <code>\%M</code>, it may fail to create a temporary file. Similarly,
* if you experience severe memory shortage, don't try to report it using
* this function, as it needs some available memory for its own internal
* working.
*
* If you did not define <b><code>NDEBUG</code></b> when compiling this
* module, after registering the event this function will call
* <a href="http://msdn.microsoft.com/library/en-us/debug/base/formatmessage.asp"
* target="_blank"><code>FormatMessage</code></a> to format the message
* and output it on <code>stdout</code> along with the return code from
* <code>ReportEvent</code>.
*
* @param log a poiter to an initialized log_event_data
* @param id the message id
* @param b_len length of the blob
* @param blob chunk of binary data
* @param args list of insertion arguments
*
* @see log_event()
* @see log_event_blob()
*/
void v_log_event(log_event_data const*log,
   DWORD id, DWORD b_len, void *blob, va_list args)
{
   parsed_fmt pf;
   char fmt[512];
   log_message_entry* lme;
   WORD severity, error_type, category, facility;
   char* printed = NULL;
   int rtc;

   if (log == NULL)
      return;

   memset(&pf, 0, sizeof pf);
   if ((lme = search_message(log, id)) != NULL)
   {
#if defined(va_copy)
      va_copy(pf.args, args);
#else
      pf.args = args;
#endif
      pf.fmt = pf.defl = &fmt[0];
      pf.fmt_len = sizeof fmt;
      fmt[0] = 0;
      reformat(&pf, lme);
      if (pf.count > MAX_INSERTIONS)
         pf.count = MAX_INSERTIONS;
      if (pf.mask_P != 0)
      {
         FILE *fp = tmpfile();
         if (fp)
         {
            int size = vfprintf(fp, fmt, pf.args);
            printed = (char*)malloc(size);
            if (printed &&
               fseek(fp, 0, SEEK_SET) == 0 &&
               fread(printed, size, 1, fp) == 1)
            {
               char *p = printed, *next;
               unsigned mask = 1 << pf.ins_count;
               while (pf.ins_count < pf.count &&
                  p - printed < size &&
                  (next = strstr(p, ARG_SEPARATOR)) != NULL)
               {
                  *next = 0;
                  pf.ins[pf.ins_count] = (pf.mask_M & mask) ?
                     alloc_sys_message(atoi(p), pf.defl) : p;
                  p = next + ARG_SEPARATOR_LENGTH;
                  mask <<= 1;
                  ++pf.ins_count;
               }
            }
            fclose(fp);
         }
      }

      strcpy(pf.defl, "DATA LOST");
      while (pf.ins_count < pf.count)
         pf.ins[pf.ins_count++] = pf.defl;
   }

   severity = (WORD)(id >> 30);
   switch (severity)
   {
      case 0:  error_type = EVENTLOG_SUCCESS; break;
      case 1:  error_type = EVENTLOG_INFORMATION_TYPE; break;
      case 2:  error_type = EVENTLOG_WARNING_TYPE; break;
      default: error_type = EVENTLOG_ERROR_TYPE; break;
   }
   facility = (WORD)((id & 0xfff0000) >> 16);
   if (log->category <= 0)
      category = facility;
   if (category > log->last_category)
      category = (WORD)log->last_category;

   if (log->log)
      rtc = ReportEvent(log->log,  // handle
         error_type,               // log type
         category,                 // event category
         id,                       // event code
         log->sid,                 // user sid
         (WORD)pf.ins_count,       // number of arguments
         b_len,                    // blob size
         &pf.ins[0],               // arguments
         blob);                    // blob
   else
      rtc = 0;

#if defined(NDEBUG)
   (void)rtc;
#else
   {
      LPVOID lpMsgBuf;
      int error = rtc ? 0 : log->log ? GetLastError() : 0;
      FormatMessage(
         FORMAT_MESSAGE_ALLOCATE_BUFFER |
         FORMAT_MESSAGE_ARGUMENT_ARRAY  |
         FORMAT_MESSAGE_FROM_HMODULE    |
         65 /* line length */ & FORMAT_MESSAGE_MAX_WIDTH_MASK,
         log->msg,
         id,
         MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language */
         (LPTSTR) &lpMsgBuf,
         0,
         (va_list*)&pf.ins[0]
      );
      printf("\n\n%s message %u (%#x) cat. %u, facility %u\n",
         rtc ? "Logged" : "FAILURE logging",
         id & 0xffff, id, category, facility);
      if (lpMsgBuf)
      {
         puts("----- Begin message -----");
         puts(lpMsgBuf);
         puts("----- End message -----\n");
         LocalFree(lpMsgBuf);
      }
      else
         printf("Message cannot be retrieved: error %d\n", GetLastError());
      if (!rtc && error)
         printf("Message cannot be logged: error %d\n", error);
   }
#endif

   if (pf.mask_M != 0)
   {
      int i;
      unsigned mask;
      for (i = 0, mask = 1; i < pf.count; ++i, mask <<= 1)
         if ((pf.mask_M & mask) != 0 && pf.ins[i] && pf.defl != pf.ins[i])
            free((void*)pf.ins[i]);
   }
   free(printed);
}

/**
* Log an event. This is coded as
* @code
*   va_list ap;
*   va_start(ap, id);
*   v_log_event(log, id, 0, NULL, ap);
*   va_end(ap);
* @endcode
*
* @param log a poiter to an initialized log_event_data
* @param id the message id
* @param ... insertion arguments
*
* @see v_log_event()
* @see log_event_blob()
*/
void log_event(log_event_data const *log, DWORD id, ...)
{
   va_list ap;
   va_start(ap, id);
   v_log_event(log, id, 0, NULL, ap);
   va_end(ap);
}

/**
* Log an event and save additional binary data.
* @param log a poiter to an initialized log_event_data
* @param id the message id
* @param blob_length length of the blob
* @param blob chunk of binary data
* @param ... insertion arguments
*
* @see v_log_event()
* @see log_event()
*/
void log_event_blob(log_event_data const*log, DWORD id,
   DWORD blob_length, void *blob, ... /* insertion strings */)
{
   va_list ap;
   va_start(ap, blob);
   v_log_event(log, id, blob_length, blob, ap);
   va_end(ap);
}

/**
* Initialize a log_event_data structure. After the structure is initialized,
* it will not be altered until you clear it, so there is no need to protect
* it from simultaneous acces by different threads.
* The caller can allocate the structure in static storage or wherever it can be
* easily accessed.
*
* Initialization fails with error 120 if not on WIN32, e.g. on win98.
* However, if NDEBUG is not defined, event will still be seen on stderr.
* In that case log_init returns 120.
*
* Pass a positive category for logging all messages with it. If the category
* is 0 or negative, the message facility will be used.
*
* Initializing an already initialized structure results in resource leaking.
*
* @param log the log_event_data structure to initialize
* @param category the category that will be used for logging
* @param pszSrcName event source name
* @param pszMsgDLL name or path for message DLL
* @return 0 if successful
*
* @see SetEventSource()
* @see GetEventSid()
* @see log_clear()
*/
int log_init(log_event_data*log, int category, LPTSTR pszSrcName, LPTSTR pszMsgDLL)
{
   HMODULE msg;
   HRSRC h;
   HGLOBAL res = NULL;
   char fname[PATH_MAX], *file;
   unsigned name_size = sizeof fname;
   log_message_table *table;
   DWORD error = 0;
   int rtc = 0;

   LOCATION_DECL;

   memset(log, 0, sizeof *log);

   file = check_full_path(pszMsgDLL, fname, &name_size);
   if (file == NULL)
   {
      error = GetLastError();
      LOCATION_SET("check_full_path");
      goto func_exit;
   }

   msg = LoadLibraryEx(file, NULL, LOAD_LIBRARY_AS_DATAFILE);
   if (msg == NULL)
   {
      error = GetLastError();
      LOCATION_SET("LoadLibraryEx");
      goto func_exit;
   }

   h = FindResourceEx(msg, RT_MESSAGETABLE, MAKEINTRESOURCE(1), 0x444);
   if (h == NULL)
   {
      error = GetLastError();
      LOCATION_SET("FindResourceEx");
      goto func_exit;
   }

   res = LoadResource(msg, h);
   if (res == NULL)
   {
      error = GetLastError();
      LOCATION_SET("LoadResource");
      goto func_exit;
   }

   table = (log_message_table*)LockResource(res);
   if (table == NULL)
   {
      error = GetLastError();
      LOCATION_SET("LockResource");
      goto func_exit;
   }

   if (table->nentries <= 0)
   {
      error = ERROR_NOT_SUPPORTED;
      LOCATION_SET("nentries");
      goto func_exit;
   }

   if ((log->log = SetEventSource(pszSrcName, file,
      log->last_category = table->index[0].last_entry)) == NULL)
   {
      error = GetLastError();
      LOCATION_SET("SetEventSource");
#if defined(NDEBUG)
      goto func_exit;
#else
      rtc = error;
      LOCATION_PRINT("log_init");
#endif
   }

   log->sid = GetEventSid();
   log->category = category;
   log->msg = msg;
   log->res = res;
   log->table = table;

   func_exit:
   {
      if (log->log == NULL && rtc == 0)
      {
         if (res)
            FreeResource(res);
         if (msg)
            FreeLibrary(msg);
         if (error)
         {
            SetLastError(error);
            LOCATION_PRINT("log_init");
         }
         log_clear(log);
      }
   }

   return rtc ? rtc : log->log == NULL; // 0 if Ok
}

/**
* Free resources and zero the memory.
* This function may be needed if you are concerned with handle inheritance
* or before exiting the program if you use a tool to detect memory leakages.
*
* @param log the log_event_data structure to be cleared
*
* @see log_init()
*/
void log_clear(log_event_data *log)
{
   if (log->msg)
      FreeLibrary(log->msg);
   if (log->res)
      FreeResource(log->res);
   if (log->sid)
      ClearEventSid(log->sid);
   if (log->log)
      DeregisterEventSource(log->log);
   memset(log, 0, sizeof *log);
}

#if defined(TEST_MAIN)
#include <msgs.h>

int main(int argc, char *argv[])
{
   int i;
   char *name = NULL, *file = NULL;
   log_event_data log;

   for (i = 1; i < argc; ++i)
   {
      if (argv[i][0] == '-' || argv[i][0] == '/' || file && name)
      {
         printf("usage:\n\t%s msg-file-name srcname\n", argv[0]);
         return 0;
      }
      if (file == NULL)
         file = argv[i];
      else if (strchr(name = argv[i], '\\') != NULL)
         --i; // error
   }
   if (name == NULL) name = "eventTest";
   if (file == NULL) file = argv[0];

   /*
   * run test even if log_init failed
   */
   if (log_init(&log, 0, name, file) == 0 || argc == 1)
   {
      double d = 3.14159265358979323846;
      __int64 i64 = 9223372036854775807i64;
      wchar_t *wide = L"Unicode string";

      log_event(&log, FIRST_MESSAGE);
      log_event(&log, MESSAGE_WITH_AN_ARGUMENT_s,
         "string argument is this");
      log_event(&log, MESSAGE_VARIOUS_ARGUMENTS_sMsasMD,
         "string argument is this",
         ERROR_NOT_ENOUGH_MEMORY,
         "another string as the second argument",
         10, "1234567890 and some more stuff...",
         ERROR_NOT_SUPPORTED,
         i64);
      log_event(&log, MESSAGE_BAD_FORMAT_s,
         "string argument is this");
      log_event(&log, MESSAGE_TOO_MANY_ARGS_ssssssssss,
         "first", "second", "third", "fourth", "fifth",
         "sixth", "seventh", "eighth", "nineth", "tenth");
      log_event(&log, MESSAGE_TYPES_diouxXfe,
         1000, 1000, 1000, 1000, 1000, 1000, d, d);
      log_event(&log, MESSAGE_TYPES_EgGcCsSp,
         d, d, d, 'U', 'U', "ansi string", wide, log_event);
      log_event(&log, MESSAGE_WRONG_COUNT_spds,
         "first", log_event, 10, "last");
      log_event(&log, MESSAGE_WRONG_COUNT_spd,
         "first", log_event, 10);

      log_event_blob(&log, FIRST_MESSAGE, sizeof log, &log);
      log_event_blob(&log, MESSAGE_WITH_AN_ARGUMENT_s, sizeof log, &log,
         "string argument is this");
      log_event_blob(&log, MESSAGE_VARIOUS_ARGUMENTS_sMsasMD, sizeof log, &log,
         "string argument is this",
         ERROR_NOT_ENOUGH_MEMORY,
         "another string as the second argument",
         10, "1234567890 and some more stuff...",
         ERROR_NOT_SUPPORTED,
         i64);
      log_event_blob(&log, MESSAGE_BAD_FORMAT_s, sizeof log, &log,
         "string argument is this");
      log_event_blob(&log, MESSAGE_TOO_MANY_ARGS_ssssssssss, sizeof log, &log,
         "first", "second", "third", "fourth", "fifth",
         "sixth", "seventh", "eighth", "nineth", "tenth");
      log_event_blob(&log, MESSAGE_TYPES_diouxXfe, sizeof log, &log,
         1000, 1000, 1000, 1000, 1000, 1000, d, d);
      log_event_blob(&log, MESSAGE_TYPES_EgGcCsSp, sizeof log, &log,
         d, d, d, 'U', 'U', "ansi string", wide, log_event);
      log_event_blob(&log, MESSAGE_WRONG_COUNT_spds, sizeof log, &log,
         "first", log_event, 10, "last");
      log_event_blob(&log, MESSAGE_WRONG_COUNT_spd, sizeof log, &log,
         "first", log_event, 10);

      log_clear(&log);
   }
   return 0;
}
#endif // TEST_MAIN
